@@ -34,11 +34,20 @@ static int
 usage(int ret)
 {
 	fprintf(stderr, "usage: ./scanner [client-header|server-header|code]\n");
+	fprintf(stderr, "\n");
+	fprintf(stderr, "Converts XML protocol descriptions supplied on "
+			"stdin to client headers,\n"
+			"server headers, or protocol marshalling code.\n");
 	exit(ret);
 }
 
 #define XML_BUFFER_SIZE 4096
 
+struct location {
+	const char *filename;
+	int line_number;
+};
+
 struct description {
 	char *summary;
 	char *text;
@@ -55,6 +64,7 @@ struct protocol {
 };
 
 struct interface {
+	struct location loc;
 	char *name;
 	char *uppercase_name;
 	int version;
@@ -67,11 +77,13 @@ struct interface {
 };
 
 struct message {
+	struct location loc;
 	char *name;
 	char *uppercase_name;
 	struct wl_list arg_list;
 	struct wl_list link;
 	int arg_count;
+	int new_id_count;
 	int type_index;
 	int all_null;
 	int destructor;
@@ -116,7 +128,7 @@ struct entry {
 };
 
 struct parse_context {
-	const char *filename;
+	struct location loc;
 	XML_Parser parser;
 	struct protocol *protocol;
 	struct interface *interface;
@@ -127,16 +139,27 @@ struct parse_context {
 	unsigned int character_data_length;
 };
 
-static int
-list_length(struct wl_list *list)
+static void *
+fail_on_null(void *p)
 {
-	struct wl_list *l;
-	int i;
+	if (p == NULL) {
+		fprintf(stderr, "wayland-scanner: out of memory\n");
+		exit(EXIT_FAILURE);
+	}
 
-	for (i = 0, l = list->next; l != list; i++, l = l->next)
-		;
+	return p;
+}
 
-	return i;
+static void *
+xmalloc(size_t s)
+{
+	return fail_on_null(malloc(s));
+}
+
+static char *
+xstrdup(const char *s)
+{
+	return fail_on_null(strdup(s));
 }
 
 static char *
@@ -145,7 +168,7 @@ uppercase_dup(const char *src)
 	char *u;
 	int i;
 
-	u = strdup(src);
+	u = xstrdup(src);
 	for (i = 0; u[i]; i++)
 		u[i] = toupper(u[i]);
 	u[i] = '\0';
@@ -234,13 +257,32 @@ desc_dump(char *desc, const char *fmt, ...)
 }
 
 static void
-fail(struct parse_context *ctx, const char *msg)
+fail(struct location *loc, const char *msg, ...)
 {
-	fprintf(stderr, "%s:%ld: %s\n",
-		ctx->filename, XML_GetCurrentLineNumber(ctx->parser), msg);
+	va_list ap;
+
+	va_start(ap, msg);
+	fprintf(stderr, "%s:%d: error: ",
+		loc->filename, loc->line_number);
+	vfprintf(stderr, msg, ap);
+	fprintf(stderr, "\n");
+	va_end(ap);
 	exit(EXIT_FAILURE);
 }
 
+static void
+warn(struct location *loc, const char *msg, ...)
+{
+	va_list ap;
+
+	va_start(ap, msg);
+	fprintf(stderr, "%s:%d: warning: ",
+		loc->filename, loc->line_number);
+	vfprintf(stderr, msg, ap);
+	fprintf(stderr, "\n");
+	va_end(ap);
+}
+
 static int
 is_nullable_type(struct arg *arg)
 {
@@ -271,6 +313,7 @@ start_element(void *data, const char *element_name, const char **atts)
 	char *end;
 	int i, version;
 
+	ctx->loc.line_number = XML_GetCurrentLineNumber(ctx->parser);
 	name = NULL;
 	type = NULL;
 	version = 0;
@@ -302,22 +345,23 @@ start_element(void *data, const char *element_name, const char **atts)
 	ctx->character_data_length = 0;
 	if (strcmp(element_name, "protocol") == 0) {
 		if (name == NULL)
-			fail(ctx, "no protocol name given");
+			fail(&ctx->loc, "no protocol name given");
 
-		ctx->protocol->name = strdup(name);
+		ctx->protocol->name = xstrdup(name);
 		ctx->protocol->uppercase_name = uppercase_dup(name);
 		ctx->protocol->description = NULL;
 	} else if (strcmp(element_name, "copyright") == 0) {
 		
 	} else if (strcmp(element_name, "interface") == 0) {
 		if (name == NULL)
-			fail(ctx, "no interface name given");
+			fail(&ctx->loc, "no interface name given");
 
 		if (version == 0)
-			fail(ctx, "no interface version given");
+			fail(&ctx->loc, "no interface version given");
 
-		interface = malloc(sizeof *interface);
-		interface->name = strdup(name);
+		interface = xmalloc(sizeof *interface);
+		interface->loc = ctx->loc;
+		interface->name = xstrdup(name);
 		interface->uppercase_name = uppercase_dup(name);
 		interface->version = version;
 		interface->description = NULL;
@@ -331,13 +375,15 @@ start_element(void *data, const char *element_name, const char **atts)
 	} else if (strcmp(element_name, "request") == 0 ||
 		   strcmp(element_name, "event") == 0) {
 		if (name == NULL)
-			fail(ctx, "no request name given");
+			fail(&ctx->loc, "no request name given");
 
-		message = malloc(sizeof *message);
-		message->name = strdup(name);
+		message = xmalloc(sizeof *message);
+		message->loc = ctx->loc;
+		message->name = xstrdup(name);
 		message->uppercase_name = uppercase_dup(name);
 		wl_list_init(&message->arg_list);
 		message->arg_count = 0;
+		message->new_id_count = 0;
 		message->description = NULL;
 
 		if (strcmp(element_name, "request") == 0)
@@ -353,26 +399,30 @@ start_element(void *data, const char *element_name, const char **atts)
 			message->destructor = 0;
 
 		if (since != NULL) {
+			errno = 0;
 			version = strtol(since, &end, 0);
 			if (errno == EINVAL || end == since || *end != '\0')
-				fail(ctx, "invalid integer\n");
-			if (version < ctx->interface->since)
-				fail(ctx, "since version not increasing\n");
-			ctx->interface->since = version;
+				fail(&ctx->loc,
+				     "invalid integer (%s)\n", since);
+		} else {
+			version = 1;
 		}
 
-		message->since = ctx->interface->since;
+		if (version < ctx->interface->since)
+			warn(&ctx->loc, "since version not increasing\n");
+		ctx->interface->since = version;
+		message->since = version;
 
 		if (strcmp(name, "destroy") == 0 && !message->destructor)
-			fail(ctx, "destroy request should be destructor type");
+			fail(&ctx->loc, "destroy request should be destructor type");
 
 		ctx->message = message;
 	} else if (strcmp(element_name, "arg") == 0) {
 		if (name == NULL)
-			fail(ctx, "no argument name given");
+			fail(&ctx->loc, "no argument name given");
 
-		arg = malloc(sizeof *arg);
-		arg->name = strdup(name);
+		arg = xmalloc(sizeof *arg);
+		arg->name = xstrdup(name);
 
 		if (strcmp(type, "int") == 0)
 			arg->type = INT;
@@ -391,20 +441,24 @@ start_element(void *data, const char *element_name, const char **atts)
 		} else if (strcmp(type, "object") == 0) {
 			arg->type = OBJECT;
 		} else {
-			fail(ctx, "unknown type");
+			fail(&ctx->loc, "unknown type (%s)", type);
 		}
 
 		switch (arg->type) {
 		case NEW_ID:
+			ctx->message->new_id_count++;
+
+			/* Fall through to OBJECT case. */
+
 		case OBJECT:
 			if (interface_name)
-				arg->interface_name = strdup(interface_name);
+				arg->interface_name = xstrdup(interface_name);
 			else
 				arg->interface_name = NULL;
 			break;
 		default:
 			if (interface_name != NULL)
-				fail(ctx, "interface not allowed");
+				fail(&ctx->loc, "interface attribute not allowed for type %s", type);
 			break;
 		}
 
@@ -413,23 +467,23 @@ start_element(void *data, const char *element_name, const char **atts)
 		else if (strcmp(allow_null, "true") == 0)
 			arg->nullable = 1;
 		else
-			fail(ctx, "invalid value for allow-null attribute");
+			fail(&ctx->loc, "invalid value for allow-null attribute (%s)", allow_null);
 
 		if (allow_null != NULL && !is_nullable_type(arg))
-			fail(ctx, "allow-null is only valid for objects, strings, and arrays");
+			fail(&ctx->loc, "allow-null is only valid for objects, strings, and arrays");
 
 		arg->summary = NULL;
 		if (summary)
-			arg->summary = strdup(summary);
+			arg->summary = xstrdup(summary);
 
 		wl_list_insert(ctx->message->arg_list.prev, &arg->link);
 		ctx->message->arg_count++;
 	} else if (strcmp(element_name, "enum") == 0) {
 		if (name == NULL)
-			fail(ctx, "no enum name given");
+			fail(&ctx->loc, "no enum name given");
 
-		enumeration = malloc(sizeof *enumeration);
-		enumeration->name = strdup(name);
+		enumeration = xmalloc(sizeof *enumeration);
+		enumeration->name = xstrdup(name);
 		enumeration->uppercase_name = uppercase_dup(name);
 		enumeration->description = NULL;
 		wl_list_init(&enumeration->entry_list);
@@ -440,24 +494,24 @@ start_element(void *data, const char *element_name, const char **atts)
 		ctx->enumeration = enumeration;
 	} else if (strcmp(element_name, "entry") == 0) {
 		if (name == NULL)
-			fail(ctx, "no entry name given");
+			fail(&ctx->loc, "no entry name given");
 
-		entry = malloc(sizeof *entry);
-		entry->name = strdup(name);
+		entry = xmalloc(sizeof *entry);
+		entry->name = xstrdup(name);
 		entry->uppercase_name = uppercase_dup(name);
-		entry->value = strdup(value);
+		entry->value = xstrdup(value);
 		if (summary)
-			entry->summary = strdup(summary);
+			entry->summary = xstrdup(summary);
 		else
 			entry->summary = NULL;
 		wl_list_insert(ctx->enumeration->entry_list.prev,
 			       &entry->link);
 	} else if (strcmp(element_name, "description") == 0) {
 		if (summary == NULL)
-			fail(ctx, "description without summary");
+			fail(&ctx->loc, "description without summary");
 
-		description = malloc(sizeof *description);
-		description->summary = strdup(summary);
+		description = xmalloc(sizeof *description);
+		description->summary = xstrdup(summary);
 
 		if (ctx->message)
 			ctx->message->description = description;
@@ -524,6 +578,18 @@ emit_opcodes(struct wl_list *message_list, struct interface *interface)
 	printf("\n");
 }
 
+static void
+emit_opcode_versions(struct wl_list *message_list, struct interface *interface)
+{
+	struct message *m;
+
+	wl_list_for_each(m, message_list, link)
+		printf("#define %s_%s_SINCE_VERSION\t%d\n",
+		       interface->uppercase_name, m->uppercase_name, m->since);
+
+	printf("\n");
+}
+
 static void
 emit_type(struct arg *a)
 {
@@ -580,18 +646,19 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 	wl_list_for_each(m, message_list, link) {
 		if (m->destructor)
 			has_destructor = 1;
-		if (strcmp(m->name, "destroy)") == 0)
+		if (strcmp(m->name, "destroy") == 0)
 			has_destroy = 1;
 	}
 
 	if (!has_destructor && has_destroy) {
-		fprintf(stderr,
-			"interface %s has method named destroy but"
-			"no destructor", interface->name);
+		fail(&interface->loc,
+		     "interface '%s' has method named destroy "
+		     "but no destructor",
+		     interface->name);
 		exit(EXIT_FAILURE);
 	}
 
-	if (!has_destructor && strcmp(interface->name, "wl_display") != 0)
+	if (!has_destroy && strcmp(interface->name, "wl_display") != 0)
 		printf("static inline void\n"
 		       "%s_destroy(struct %s *%s)\n"
 		       "{\n"
@@ -605,6 +672,14 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 		return;
 
 	wl_list_for_each(m, message_list, link) {
+		if (m->new_id_count > 1) {
+			warn(&m->loc,
+			     "request '%s::%s' has more than "
+			     "one new_id arg, not emitting stub\n",
+			     interface->name, m->name);
+			continue;
+		}
+
 		ret = NULL;
 		wl_list_for_each(a, &m->arg_list, link) {
 			if (a->type == NEW_ID)
@@ -639,31 +714,34 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 		       "{\n");
 		if (ret) {
 			printf("\tstruct wl_proxy *%s;\n\n"
-			       "\t%s = wl_proxy_create("
-			       "(struct wl_proxy *) %s,\n",
-			       ret->name, ret->name, interface->name);
-			if (ret->interface_name == NULL)
-				printf("\t\t\t     interface);\n");
-			else
-				printf("\t\t\t     &%s_interface);\n",
-				       ret->interface_name);
+			       "\t%s = wl_proxy_marshal_constructor("
+			       "(struct wl_proxy *) %s,\n"
+			       "\t\t\t %s_%s, ",
+			       ret->name, ret->name,
+			       interface->name,
+			       interface->uppercase_name,
+			       m->uppercase_name);
 
-			printf("\tif (!%s)\n"
-			       "\t\treturn NULL;\n\n",
-			       ret->name);
+			if (ret->interface_name == NULL)
+				printf("interface");
+			else
+				printf("&%s_interface", ret->interface_name);
+		} else {
+			printf("\twl_proxy_marshal((struct wl_proxy *) %s,\n"
+			       "\t\t\t %s_%s",
+			       interface->name,
+			       interface->uppercase_name,
+			       m->uppercase_name);
 		}
 
-		printf("\twl_proxy_marshal((struct wl_proxy *) %s,\n"
-		       "\t\t\t %s_%s",
-		       interface->name,
-		       interface->uppercase_name,
-		       m->uppercase_name);
-
 		wl_list_for_each(a, &m->arg_list, link) {
-			if (a->type == NEW_ID && a->interface_name == NULL)
-				printf(", interface->name, version");
-			printf(", ");
-			printf("%s", a->name);
+			if (a->type == NEW_ID) {
+				if (a->interface_name == NULL)
+					printf(", interface->name, version");
+				printf(", NULL");
+			} else {
+				printf(", %s", a->name);
+			}
 		}
 		printf(");\n");
 
@@ -841,7 +919,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 			else if (is_interface && a->type == NEW_ID && a->interface_name == NULL)
 				printf("const char *interface, uint32_t version, uint32_t ");
 			else if (!is_interface && a->type == OBJECT && a->interface_name == NULL)
-				printf("struct wl_object *");
+				printf("void *");
 
 			else if (!is_interface && a->type == NEW_ID)
 				printf("struct %s *", a->interface_name);
@@ -938,6 +1016,7 @@ emit_header(struct protocol *protocol, int server)
 		if (server) {
 			emit_structs(&i->request_list, i);
 			emit_opcodes(&i->event_list, i);
+			emit_opcode_versions(&i->event_list, i);
 			emit_event_wrappers(&i->event_list, i);
 		} else {
 			emit_structs(&i->event_list, i);
@@ -955,11 +1034,13 @@ emit_header(struct protocol *protocol, int server)
 
 static void
 emit_types_forward_declarations(struct protocol *protocol,
-				struct wl_list *message_list)
+				struct wl_list *message_list,
+				struct wl_array *types)
 {
 	struct message *m;
 	struct arg *a;
 	int length;
+	char **p;
 
 	wl_list_for_each(m, message_list, link) {
 		length = 0;
@@ -973,8 +1054,8 @@ emit_types_forward_declarations(struct protocol *protocol,
 					continue;
 
 				m->all_null = 0;
-				printf("extern const struct wl_interface %s_interface;\n",
-				       a->interface_name);
+				p = fail_on_null(wl_array_add(types, sizeof *p));
+				*p = a->interface_name;
 				break;
 			default:
 				break;
@@ -1045,6 +1126,10 @@ emit_messages(struct wl_list *message_list,
 
 	wl_list_for_each(m, message_list, link) {
 		printf("\t{ \"%s\", \"", m->name);
+
+		if (m->since > 1)
+			printf("%d", m->since);
+
 		wl_list_for_each(a, &m->arg_list, link) {
 			if (is_nullable_type(a) && a->nullable)
 				printf("?");
@@ -1085,10 +1170,20 @@ emit_messages(struct wl_list *message_list,
 	printf("};\n\n");
 }
 
+static int
+cmp_names(const void *p1, const void *p2)
+{
+	const char * const *s1 = p1, * const *s2 = p2;
+
+	return strcmp(*s1, *s2);
+}
+
 static void
 emit_code(struct protocol *protocol)
 {
 	struct interface *i;
+	struct wl_array types;
+	char **p, *prev;
 
 	if (protocol->copyright)
 		format_copyright(protocol->copyright);
@@ -1097,10 +1192,20 @@ emit_code(struct protocol *protocol)
 	       "#include <stdint.h>\n"
 	       "#include \"wayland-util.h\"\n\n");
 
+	wl_array_init(&types);
 	wl_list_for_each(i, &protocol->interface_list, link) {
-		emit_types_forward_declarations(protocol, &i->request_list);
-		emit_types_forward_declarations(protocol, &i->event_list);
+		emit_types_forward_declarations(protocol, &i->request_list, &types);
+		emit_types_forward_declarations(protocol, &i->event_list, &types);
 	}
+	qsort(types.data, types.size / sizeof *p, sizeof *p, cmp_names);
+	prev = NULL;
+	wl_array_for_each(p, &types) {
+		if (prev && strcmp(*p, prev) == 0)
+			continue;
+		printf("extern const struct wl_interface %s_interface;\n", *p);
+		prev = *p;
+	}
+	wl_array_release(&types);
 	printf("\n");
 
 	printf("static const struct wl_interface *types[] = {\n");
@@ -1123,13 +1228,13 @@ emit_code(struct protocol *protocol)
 
 		if (!wl_list_empty(&i->request_list))
 			printf("\t%d, %s_requests,\n",
-			       list_length(&i->request_list), i->name);
+			       wl_list_length(&i->request_list), i->name);
 		else
 			printf("\t0, NULL,\n");
 
 		if (!wl_list_empty(&i->event_list))
 			printf("\t%d, %s_events,\n",
-			       list_length(&i->event_list), i->name);
+			       wl_list_length(&i->event_list), i->name);
 		else
 			printf("\t0, NULL,\n");
 
@@ -1143,9 +1248,24 @@ int main(int argc, char *argv[])
 	struct protocol protocol;
 	int len;
 	void *buf;
+	enum {
+		CLIENT_HEADER,
+		SERVER_HEADER,
+		CODE,
+	} mode;
 
 	if (argc != 2)
 		usage(EXIT_FAILURE);
+	else if (strcmp(argv[1], "help") == 0 || strcmp(argv[1], "--help") == 0)
+		usage(EXIT_SUCCESS);
+	else if (strcmp(argv[1], "client-header") == 0)
+		mode = CLIENT_HEADER;
+	else if (strcmp(argv[1], "server-header") == 0)
+		mode = SERVER_HEADER;
+	else if (strcmp(argv[1], "code") == 0)
+		mode = CODE;
+	else
+		usage(EXIT_FAILURE);
 
 	wl_list_init(&protocol.interface_list);
 	protocol.type_index = 0;
@@ -1154,7 +1274,7 @@ int main(int argc, char *argv[])
 	memset(&ctx, 0, sizeof ctx);
 	ctx.protocol = &protocol;
 
-	ctx.filename = "<stdin>";
+	ctx.loc.filename = "<stdin>";
 	ctx.parser = XML_ParserCreate(NULL);
 	XML_SetUserData(ctx.parser, &ctx);
 	if (ctx.parser == NULL) {
@@ -1178,12 +1298,16 @@ int main(int argc, char *argv[])
 
 	XML_ParserFree(ctx.parser);
 
-	if (strcmp(argv[1], "client-header") == 0) {
-		emit_header(&protocol, 0);
-	} else if (strcmp(argv[1], "server-header") == 0) {
-		emit_header(&protocol, 1);
-	} else if (strcmp(argv[1], "code") == 0) {
-		emit_code(&protocol);
+	switch (mode) {
+		case CLIENT_HEADER:
+			emit_header(&protocol, 0);
+			break;
+		case SERVER_HEADER:
+			emit_header(&protocol, 1);
+			break;
+		case CODE:
+			emit_code(&protocol);
+			break;
 	}
 
 	return 0;
