@@ -21,6 +21,8 @@
  * OF THIS SOFTWARE.
  */
 
+#define _GNU_SOURCE
+
 #include <stdlib.h>
 #include <stdint.h>
 #include <stddef.h>
@@ -34,7 +36,7 @@
 #include <ctype.h>
 #include <assert.h>
 #include <fcntl.h>
-#include <sys/poll.h>
+#include <poll.h>
 #include <pthread.h>
 
 #include "wayland-util.h"
@@ -42,7 +44,6 @@
 #include "wayland-client.h"
 #include "wayland-private.h"
 
-
 /** \cond */
 
 enum wl_proxy_flag {
@@ -57,6 +58,7 @@ struct wl_proxy {
 	uint32_t flags;
 	int refcount;
 	void *user_data;
+	wl_dispatcher_func_t dispatcher;
 };
 
 struct wl_global {
@@ -80,14 +82,19 @@ struct wl_display {
 	int fd;
 	pthread_t display_thread;
 	struct wl_map objects;
-	struct wl_event_queue queue;
+	struct wl_event_queue display_queue;
+	struct wl_event_queue default_queue;
 	struct wl_list event_queue_list;
 	pthread_mutex_t mutex;
+
+	int reader_count;
+	uint32_t read_serial;
+	pthread_cond_t reader_cond;
 };
 
 /** \endcond */
 
-static int wl_debug = 0;
+static int debug_client = 0;
 
 static void
 display_fatal_error(struct wl_display *display, int error)
@@ -101,8 +108,6 @@ display_fatal_error(struct wl_display *display, int error)
 		error = 1;
 
 	display->last_error = error;
-	close(display->fd);
-	display->fd = -1;
 
 	wl_list_for_each(iter, &display->event_queue_list, link)
 		pthread_cond_broadcast(&iter->cond);
@@ -189,6 +194,29 @@ wl_display_create_queue(struct wl_display *display)
 	return queue;
 }
 
+static struct wl_proxy *
+proxy_create(struct wl_proxy *factory, const struct wl_interface *interface)
+{
+	struct wl_proxy *proxy;
+	struct wl_display *display = factory->display;
+
+	proxy = malloc(sizeof *proxy);
+	if (proxy == NULL)
+		return NULL;
+
+	proxy->object.interface = interface;
+	proxy->object.implementation = NULL;
+	proxy->dispatcher = NULL;
+	proxy->display = display;
+	proxy->queue = factory->queue;
+	proxy->flags = 0;
+	proxy->refcount = 1;
+
+	proxy->object.id = wl_map_insert_new(&display->objects, 0, proxy);
+
+	return proxy;
+}
+
 /** Create a proxy object with a given interface
  *
  * \param factory Factory proxy object
@@ -211,23 +239,11 @@ wl_display_create_queue(struct wl_display *display)
 WL_EXPORT struct wl_proxy *
 wl_proxy_create(struct wl_proxy *factory, const struct wl_interface *interface)
 {
-	struct wl_proxy *proxy;
 	struct wl_display *display = factory->display;
-
-	proxy = malloc(sizeof *proxy);
-	if (proxy == NULL)
-		return NULL;
-
-	proxy->object.interface = interface;
-	proxy->object.implementation = NULL;
-	proxy->display = display;
-	proxy->queue = factory->queue;
-	proxy->flags = 0;
-	proxy->refcount = 1;
+	struct wl_proxy *proxy;
 
 	pthread_mutex_lock(&display->mutex);
-	proxy->object.id = wl_map_insert_new(&display->objects,
-					     WL_MAP_CLIENT_SIDE, proxy);
+	proxy = proxy_create(factory, interface);
 	pthread_mutex_unlock(&display->mutex);
 
 	return proxy;
@@ -248,12 +264,13 @@ wl_proxy_create_for_id(struct wl_proxy *factory,
 	proxy->object.interface = interface;
 	proxy->object.implementation = NULL;
 	proxy->object.id = id;
+	proxy->dispatcher = NULL;
 	proxy->display = display;
 	proxy->queue = factory->queue;
 	proxy->flags = 0;
 	proxy->refcount = 1;
 
-	wl_map_insert_at(&display->objects, id, proxy);
+	wl_map_insert_at(&display->objects, 0, id, proxy);
 
 	return proxy;
 }
@@ -274,10 +291,10 @@ wl_proxy_destroy(struct wl_proxy *proxy)
 	if (proxy->flags & WL_PROXY_FLAG_ID_DELETED)
 		wl_map_remove(&proxy->display->objects, proxy->object.id);
 	else if (proxy->object.id < WL_SERVER_ID_START)
-		wl_map_insert_at(&proxy->display->objects,
+		wl_map_insert_at(&proxy->display->objects, 0,
 				 proxy->object.id, WL_ZOMBIE_OBJECT);
 	else
-		wl_map_insert_at(&proxy->display->objects,
+		wl_map_insert_at(&proxy->display->objects, 0,
 				 proxy->object.id, NULL);
 
 
@@ -311,8 +328,8 @@ WL_EXPORT int
 wl_proxy_add_listener(struct wl_proxy *proxy,
 		      void (**implementation)(void), void *data)
 {
-	if (proxy->object.implementation) {
-		fprintf(stderr, "proxy already has listener\n");
+	if (proxy->object.implementation || proxy->dispatcher) {
+		wl_log("proxy %p already has listener\n", proxy);
 		return -1;
 	}
 
@@ -322,27 +339,161 @@ wl_proxy_add_listener(struct wl_proxy *proxy,
 	return 0;
 }
 
+/** Get a proxy's listener
+ *
+ * \param proxy The proxy object
+ * \return The address of the proxy's listener or NULL if no listener is set
+ *
+ * Gets the address to the proxy's listener; which is the listener set with
+ * \ref wl_proxy_add_listener.
+ *
+ * This function is useful in client with multiple listeners on the same
+ * interface to allow the identification of which code to eexecute.
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT const void *
+wl_proxy_get_listener(struct wl_proxy *proxy)
+{
+	return proxy->object.implementation;
+}
+
+/** Set a proxy's listener (with dispatcher)
+ *
+ * \param proxy The proxy object
+ * \param dispatcher The dispatcher to be used for this proxy
+ * \param implementation The dispatcher-specific listener implementation
+ * \param data User data to be associated with the proxy
+ * \return 0 on success or -1 on failure
+ *
+ * Set proxy's listener to use \c dispatcher_func as its dispatcher and \c
+ * dispatcher_data as its dispatcher-specific implementation and its user data
+ * to \c data. If a listener has already been set, this function
+ * fails and nothing is changed.
+ *
+ * The exact details of dispatcher_data depend on the dispatcher used.  This
+ * function is intended to be used by language bindings, not user code.
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT int
+wl_proxy_add_dispatcher(struct wl_proxy *proxy,
+			wl_dispatcher_func_t dispatcher,
+			const void *implementation, void *data)
+{
+	if (proxy->object.implementation || proxy->dispatcher) {
+		wl_log("proxy %p already has listener\n");
+		return -1;
+	}
+
+	proxy->object.implementation = implementation;
+	proxy->dispatcher = dispatcher;
+	proxy->user_data = data;
+
+	return 0;
+}
+
+static struct wl_proxy *
+create_outgoing_proxy(struct wl_proxy *proxy, const struct wl_message *message,
+		      union wl_argument *args,
+		      const struct wl_interface *interface)
+{
+	int i, count;
+	const char *signature;
+	struct argument_details arg;
+	struct wl_proxy *new_proxy = NULL;
+
+	signature = message->signature;
+	count = arg_count_for_signature(signature);
+	for (i = 0; i < count; i++) {
+		signature = get_next_argument(signature, &arg);
+
+		switch (arg.type) {
+		case 'n':
+			new_proxy = proxy_create(proxy, interface);
+			if (new_proxy == NULL)
+				return NULL;
+
+			args[i].o = &new_proxy->object;
+			break;
+		}
+	}
+
+	return new_proxy;
+}
+
+/** Prepare a request to be sent to the compositor
+ *
+ * \param proxy The proxy object
+ * \param opcode Opcode of the request to be sent
+ * \param args Extra arguments for the given request
+ * \param interface The interface to use for the new proxy
+ *
+ * Translates the request given by opcode and the extra arguments into the
+ * wire format and write it to the connection buffer.  This version takes an
+ * array of the union type wl_argument.
+ *
+ * For new-id arguments, this function will allocate a new wl_proxy
+ * and send the ID to the server.  The new wl_proxy will be returned
+ * on success or NULL on errror with errno set accordingly.
+ *
+ * \note This is intended to be used by language bindings and not in
+ * non-generated code.
+ *
+ * \sa wl_proxy_marshal()
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT struct wl_proxy *
+wl_proxy_marshal_array_constructor(struct wl_proxy *proxy,
+				   uint32_t opcode, union wl_argument *args,
+				   const struct wl_interface *interface)
+{
+	struct wl_closure *closure;
+	struct wl_proxy *new_proxy = NULL;
+	const struct wl_message *message;
+
+	pthread_mutex_lock(&proxy->display->mutex);
+
+	message = &proxy->object.interface->methods[opcode];
+	if (interface) {
+		new_proxy = create_outgoing_proxy(proxy, message,
+						  args, interface);
+		if (new_proxy == NULL)
+			goto err_unlock;
+	}
+
+	closure = wl_closure_marshal(&proxy->object, opcode, args, message);
+	if (closure == NULL) {
+		wl_log("Error marshalling request: %m\n");
+		abort();
+	}
+
+	if (debug_client)
+		wl_closure_print(closure, &proxy->object, true);
+
+	if (wl_closure_send(closure, proxy->display->connection)) {
+		wl_log("Error sending request: %m\n");
+		abort();
+	}
+
+	wl_closure_destroy(closure);
+
+ err_unlock:
+	pthread_mutex_unlock(&proxy->display->mutex);
+
+	return new_proxy;
+}
+
+
 /** Prepare a request to be sent to the compositor
  *
  * \param proxy The proxy object
  * \param opcode Opcode of the request to be sent
  * \param ... Extra arguments for the given request
  *
- * Translates the request given by opcode and the extra arguments into the
- * wire format and write it to the connection buffer.
- *
- * The example below creates a proxy object with the wl_surface_interface
- * using a wl_compositor factory interface and sends the
- * \c compositor.create_surface request using \ref wl_proxy_marshal(). Note
- * the \c id is the extra argument to the request as specified by the
- * protocol.
- *
- * \code
- * id = wl_proxy_create((struct wl_proxy *) wl_compositor,
- *                      &wl_surface_interface);
- * wl_proxy_marshal((struct wl_proxy *) wl_compositor,
- *                  WL_COMPOSITOR_CREATE_SURFACE, id);
- * \endcode
+ * This function is similar to wl_proxy_marshal_constructor(), except
+ * it doesn't create proxies for new-id arguments.
  *
  * \note This should not normally be used by non-generated code.
  *
@@ -353,54 +504,96 @@ wl_proxy_add_listener(struct wl_proxy *proxy,
 WL_EXPORT void
 wl_proxy_marshal(struct wl_proxy *proxy, uint32_t opcode, ...)
 {
-	struct wl_closure *closure;
+	union wl_argument args[WL_CLOSURE_MAX_ARGS];
 	va_list ap;
 
-	pthread_mutex_lock(&proxy->display->mutex);
-
 	va_start(ap, opcode);
-	closure = wl_closure_vmarshal(&proxy->object, opcode, ap,
-				      &proxy->object.interface->methods[opcode]);
+	wl_argument_from_va_list(proxy->object.interface->methods[opcode].signature,
+				 args, WL_CLOSURE_MAX_ARGS, ap);
 	va_end(ap);
 
-	if (closure == NULL) {
-		fprintf(stderr, "Error marshalling request\n");
-		abort();
-	}
+	wl_proxy_marshal_array_constructor(proxy, opcode, args, NULL);
+}
 
-	if (wl_debug)
-		wl_closure_print(closure, &proxy->object, true);
+/** Prepare a request to be sent to the compositor
+ *
+ * \param proxy The proxy object
+ * \param opcode Opcode of the request to be sent
+ * \param interface The interface to use for the new proxy
+ * \param ... Extra arguments for the given request
+ * \return A new wl_proxy for the new_id argument or NULL on error
+ *
+ * Translates the request given by opcode and the extra arguments into the
+ * wire format and write it to the connection buffer.
+ *
+ * For new-id arguments, this function will allocate a new wl_proxy
+ * and send the ID to the server.  The new wl_proxy will be returned
+ * on success or NULL on errror with errno set accordingly.
+ *
+ * \note This should not normally be used by non-generated code.
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT struct wl_proxy *
+wl_proxy_marshal_constructor(struct wl_proxy *proxy, uint32_t opcode,
+			     const struct wl_interface *interface, ...)
+{
+	union wl_argument args[WL_CLOSURE_MAX_ARGS];
+	va_list ap;
 
-	if (wl_closure_send(closure, proxy->display->connection)) {
-		fprintf(stderr, "Error sending request: %m\n");
-		abort();
-	}
+	va_start(ap, interface);
+	wl_argument_from_va_list(proxy->object.interface->methods[opcode].signature,
+				 args, WL_CLOSURE_MAX_ARGS, ap);
+	va_end(ap);
 
-	wl_closure_destroy(closure);
+	return wl_proxy_marshal_array_constructor(proxy, opcode,
+						  args, interface);
+}
 
-	pthread_mutex_unlock(&proxy->display->mutex);
+/** Prepare a request to be sent to the compositor
+ *
+ * \param proxy The proxy object
+ * \param opcode Opcode of the request to be sent
+ * \param args Extra arguments for the given request
+ *
+ * This function is similar to wl_proxy_marshal_array_constructor(), except
+ * it doesn't create proxies for new-id arguments.
+ *
+ * \note This is intended to be used by language bindings and not in
+ * non-generated code.
+ *
+ * \sa wl_proxy_marshal()
+ *
+ * \memberof wl_proxy
+ */
+WL_EXPORT void
+wl_proxy_marshal_array(struct wl_proxy *proxy, uint32_t opcode,
+		       union wl_argument *args)
+{
+	wl_proxy_marshal_array_constructor(proxy, opcode, args, NULL);
 }
 
 static void
 display_handle_error(void *data,
-		     struct wl_display *display, struct wl_object *object,
+		     struct wl_display *display, void *object,
 		     uint32_t code, const char *message)
 {
+	struct wl_proxy *proxy = object;
 	int err;
 
 	wl_log("%s@%u: error %d: %s\n",
-	       object->interface->name, object->id, code, message);
+	       proxy->object.interface->name, proxy->object.id, code, message);
 
 	switch (code) {
 	case WL_DISPLAY_ERROR_INVALID_OBJECT:
 	case WL_DISPLAY_ERROR_INVALID_METHOD:
-		err = -EINVAL;
+		err = EINVAL;
 		break;
 	case WL_DISPLAY_ERROR_NO_MEMORY:
-		err = -ENOMEM;
+		err = ENOMEM;
 		break;
 	default:
-		err = -EFAULT;
+		err = EFAULT;
 		break;
 	}
 
@@ -442,9 +635,7 @@ connect_to_socket(const char *name)
 
 	runtime_dir = getenv("XDG_RUNTIME_DIR");
 	if (!runtime_dir) {
-		fprintf(stderr,
-			"error: XDG_RUNTIME_DIR not set in the environment.\n");
-
+		wl_log("error: XDG_RUNTIME_DIR not set in the environment.\n");
 		/* to prevent programs reporting
 		 * "failed to create display: Success" */
 		errno = ENOENT;
@@ -468,8 +659,7 @@ connect_to_socket(const char *name)
 
 	assert(name_size > 0);
 	if (name_size > (int)sizeof addr.sun_path) {
-		fprintf(stderr,
-		       "error: socket path \"%s/%s\" plus null terminator"
+		wl_log("error: socket path \"%s/%s\" plus null terminator"
 		       " exceeds 108 bytes\n", runtime_dir, name);
 		close(fd);
 		/* to prevent programs reporting
@@ -507,7 +697,7 @@ wl_display_connect_to_fd(int fd)
 
 	debug = getenv("WAYLAND_DEBUG");
 	if (debug && (strstr(debug, "client") || strstr(debug, "1")))
-		wl_debug = 1;
+		debug_client = 1;
 
 	display = malloc(sizeof *display);
 	if (display == NULL) {
@@ -518,33 +708,40 @@ wl_display_connect_to_fd(int fd)
 	memset(display, 0, sizeof *display);
 
 	display->fd = fd;
-	wl_map_init(&display->objects);
-	wl_event_queue_init(&display->queue, display);
+	wl_map_init(&display->objects, WL_MAP_CLIENT_SIDE);
+	wl_event_queue_init(&display->default_queue, display);
+	wl_event_queue_init(&display->display_queue, display);
 	wl_list_init(&display->event_queue_list);
 	pthread_mutex_init(&display->mutex, NULL);
+	pthread_cond_init(&display->reader_cond, NULL);
+	display->reader_count = 0;
 
-	wl_map_insert_new(&display->objects, WL_MAP_CLIENT_SIDE, NULL);
+	wl_map_insert_new(&display->objects, 0, NULL);
 
 	display->proxy.object.interface = &wl_display_interface;
 	display->proxy.object.id =
-		wl_map_insert_new(&display->objects,
-				  WL_MAP_CLIENT_SIDE, display);
+		wl_map_insert_new(&display->objects, 0, display);
 	display->proxy.display = display;
 	display->proxy.object.implementation = (void(**)(void)) &display_listener;
 	display->proxy.user_data = display;
-	display->proxy.queue = &display->queue;
+	display->proxy.queue = &display->default_queue;
 	display->proxy.flags = 0;
 	display->proxy.refcount = 1;
 
 	display->connection = wl_connection_create(display->fd);
-	if (display->connection == NULL) {
-		wl_map_release(&display->objects);
-		close(display->fd);
-		free(display);
-		return NULL;
-	}
+	if (display->connection == NULL)
+		goto err_connection;
 
 	return display;
+
+ err_connection:
+	pthread_mutex_destroy(&display->mutex);
+	pthread_cond_destroy(&display->reader_cond);
+	wl_map_release(&display->objects);
+	close(display->fd);
+	free(display);
+
+	return NULL;
 }
 
 /** Connect to a Wayland display
@@ -553,7 +750,7 @@ wl_display_connect_to_fd(int fd)
  * \return A \ref wl_display object or \c NULL on failure
  *
  * Connect to the Wayland display named \c name. If \c name is \c NULL,
- * its value will bee replaced with the WAYLAND_DISPLAY environment
+ * its value will be replaced with the WAYLAND_DISPLAY environment
  * variable if it is set, otherwise display "wayland-0" will be used.
  *
  * \memberof wl_display
@@ -597,10 +794,10 @@ wl_display_disconnect(struct wl_display *display)
 {
 	wl_connection_destroy(display->connection);
 	wl_map_release(&display->objects);
-	wl_event_queue_release(&display->queue);
+	wl_event_queue_release(&display->default_queue);
 	pthread_mutex_destroy(&display->mutex);
-	if (display->fd > 0)
-		close(display->fd);
+	pthread_cond_destroy(&display->reader_cond);
+	close(display->fd);
 
 	free(display);
 }
@@ -652,6 +849,8 @@ wl_display_roundtrip(struct wl_display *display)
 
 	done = 0;
 	callback = wl_display_sync(display);
+	if (callback == NULL)
+		return -1;
 	wl_callback_add_listener(callback, &sync_listener, &done);
 	while (!done && ret >= 0)
 		ret = wl_display_dispatch(display);
@@ -730,6 +929,7 @@ queue_event(struct wl_display *display, int len)
 	struct wl_proxy *proxy;
 	struct wl_closure *closure;
 	const struct wl_message *message;
+	struct wl_event_queue *queue;
 
 	wl_connection_copy(display->connection, p, sizeof p);
 	id = p[0];
@@ -767,9 +967,14 @@ queue_event(struct wl_display *display, int len)
 	proxy->refcount++;
 	closure->proxy = proxy;
 
-	if (wl_list_empty(&proxy->queue->event_list))
-		pthread_cond_signal(&proxy->queue->cond);
-	wl_list_insert(proxy->queue->event_list.prev, &closure->link);
+	if (proxy == &display->proxy)
+		queue = &display->display_queue;
+	else
+		queue = proxy->queue;
+
+	if (wl_list_empty(&queue->event_list))
+		pthread_cond_signal(&queue->cond);
+	wl_list_insert(queue->event_list.prev, &closure->link);
 
 	return size;
 }
@@ -836,13 +1041,18 @@ dispatch_event(struct wl_display *display, struct wl_event_queue *queue)
 
 	pthread_mutex_unlock(&display->mutex);
 
-	if (proxy->object.implementation) {
-		if (wl_debug)
+	if (proxy->dispatcher) {
+		if (debug_client)
+			wl_closure_print(closure, &proxy->object, false);
+
+		wl_closure_dispatch(closure, proxy->dispatcher,
+				    &proxy->object, opcode);
+	} else if (proxy->object.implementation) {
+		if (debug_client)
 			wl_closure_print(closure, &proxy->object, false);
 
 		wl_closure_invoke(closure, WL_CLOSURE_INVOKE_CLIENT,
-				  &proxy->object, opcode,
-				  proxy->user_data);
+				  &proxy->object, opcode, proxy->user_data);
 	}
 
 	wl_closure_destroy(closure);
@@ -851,65 +1061,226 @@ dispatch_event(struct wl_display *display, struct wl_event_queue *queue)
 }
 
 static int
-dispatch_queue(struct wl_display *display,
-	       struct wl_event_queue *queue, int block)
+read_events(struct wl_display *display)
 {
-	int len, size, count, ret;
+	int total, rem, size;
+	uint32_t serial;
 
-	pthread_mutex_lock(&display->mutex);
+	display->reader_count--;
+	if (display->reader_count == 0) {
+		total = wl_connection_read(display->connection);
+		if (total == -1) {
+			if (errno == EAGAIN)
+				return 0;
 
-	if (display->last_error)
-		goto err_unlock;
-
-	ret = wl_connection_flush(display->connection);
-	if (ret < 0 && errno != EAGAIN) {
-		display_fatal_error(display, errno);
-		goto err_unlock;
-	}
-
-	if (block && wl_list_empty(&queue->event_list) &&
-	    pthread_equal(display->display_thread, pthread_self())) {
-		len = wl_connection_read(display->connection);
-		if (len == -1) {
 			display_fatal_error(display, errno);
-			goto err_unlock;
-		} else if (len == 0) {
-			display_fatal_error(display, EPIPE);
-			goto err_unlock;
+			return -1;
+		} else if (total == 0) {
+			/* The compositor has closed the socket. This
+			 * should be considered an error so we'll fake
+			 * an errno */
+			errno = EPIPE;
+			display_fatal_error(display, errno);
+			return -1;
 		}
-		while (len >= 8) {
-			size = queue_event(display, len);
+
+		for (rem = total; rem >= 8; rem -= size) {
+			size = queue_event(display, rem);
 			if (size == -1) {
 				display_fatal_error(display, errno);
-				goto err_unlock;
+				return -1;
 			} else if (size == 0) {
 				break;
 			}
-			len -= size;
 		}
-	} else if (block && wl_list_empty(&queue->event_list)) {
-		pthread_cond_wait(&queue->cond, &display->mutex);
-		if (display->last_error)
-			goto err_unlock;
+
+		display->read_serial++;
+		pthread_cond_broadcast(&display->reader_cond);
+	} else {
+		serial = display->read_serial;
+		while (display->read_serial == serial)
+			pthread_cond_wait(&display->reader_cond,
+					  &display->mutex);
 	}
 
-	for (count = 0; !wl_list_empty(&queue->event_list); count++) {
-		dispatch_event(display, queue);
-		if (display->last_error)
-			goto err_unlock;
-	}
+	return 0;
+}
+
+/** Read events from display file descriptor
+ *
+ * \param display The display context object
+ * \return 0 on success or -1 on error.  In case of error errno will
+ * be set accordingly
+ *
+ * This will read events from the file descriptor for the display.
+ * This function does not dispatch events, it only reads and queues
+ * events into their corresponding event queues.  If no data is
+ * avilable on the file descriptor, wl_display_read_events() returns
+ * immediately.  To dispatch events that may have been queued, call
+ * wl_display_dispatch_pending() or
+ * wl_display_dispatch_queue_pending().
+ *
+ * Before calling this function, wl_display_prepare_read() must be
+ * called first.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT int
+wl_display_read_events(struct wl_display *display)
+{
+	int ret;
+
+	pthread_mutex_lock(&display->mutex);
+
+	ret = read_events(display);
 
 	pthread_mutex_unlock(&display->mutex);
 
+	return ret;
+}
+
+static int
+dispatch_queue(struct wl_display *display, struct wl_event_queue *queue)
+{
+	int count;
+
+	if (display->last_error)
+		goto err;
+
+	count = 0;
+	while (!wl_list_empty(&display->display_queue.event_list)) {
+		dispatch_event(display, &display->display_queue);
+		if (display->last_error)
+			goto err;
+		count++;
+	}
+
+	while (!wl_list_empty(&queue->event_list)) {
+		dispatch_event(display, queue);
+		if (display->last_error)
+			goto err;
+		count++;
+	}
+
 	return count;
 
-err_unlock:
+err:
 	errno = display->last_error;
-	pthread_mutex_unlock(&display->mutex);
 
 	return -1;
 }
 
+WL_EXPORT int
+wl_display_prepare_read_queue(struct wl_display *display,
+			      struct wl_event_queue *queue)
+{
+	int ret;
+
+	pthread_mutex_lock(&display->mutex);
+
+	if (!wl_list_empty(&queue->event_list)) {
+		errno = EAGAIN;
+		ret = -1;
+	} else {
+		display->reader_count++;
+		ret = 0;
+	}
+
+	pthread_mutex_unlock(&display->mutex);
+
+	return ret;
+}
+
+/** Prepare to read events after polling file descriptor
+ *
+ * \param display The display context object
+ * \return 0 on success or -1 if event queue was not empty
+ *
+ * This function must be called before reading from the file
+ * descriptor using wl_display_read_events().  Calling
+ * wl_display_prepare_read() announces the calling threads intention
+ * to read and ensures that until the thread is ready to read and
+ * calls wl_display_read_events(), no other thread will read from the
+ * file descriptor.  This only succeeds if the event queue is empty
+ * though, and if there are undispatched events in the queue, -1 is
+ * returned and errno set to EAGAIN.
+ *
+ * If a thread successfully calls wl_display_prepare_read(), it must
+ * either call wl_display_read_events() when it's ready or cancel the
+ * read intention by calling wl_display_cancel_read().
+ *
+ * Use this function before polling on the display fd or to integrate
+ * the fd into a toolkit event loop in a race-free way.  Typically, a
+ * toolkit will call wl_display_dispatch_pending() before sleeping, to
+ * make sure it doesn't block with unhandled events.  Upon waking up,
+ * it will assume the file descriptor is readable and read events from
+ * the fd by calling wl_display_dispatch().  Simplified, we have:
+ *
+ *   wl_display_dispatch_pending(display);
+ *   wl_display_flush(display);
+ *   poll(fds, nfds, -1);
+ *   wl_display_dispatch(display);
+ *
+ * There are two races here: first, before blocking in poll(), the fd
+ * could become readable and another thread reads the events.  Some of
+ * these events may be for the main queue and the other thread will
+ * queue them there and then the main thread will go to sleep in
+ * poll().  This will stall the application, which could be waiting
+ * for a event to kick of the next animation frame, for example.
+ *
+ * The other race is immediately after poll(), where another thread
+ * could preempt and read events before the main thread calls
+ * wl_display_dispatch().  This call now blocks and starves the other
+ * fds in the event loop.
+ *
+ * A correct sequence would be:
+ *
+ *   while (wl_display_prepare_read(display) != 0)
+ *           wl_display_dispatch_pending(display);
+ *   wl_display_flush(display);
+ *   poll(fds, nfds, -1);
+ *   wl_display_read_events(display);
+ *   wl_display_dispatch_pending(display);
+ *
+ * Here we call wl_display_prepare_read(), which ensures that between
+ * returning from that call and eventually calling
+ * wl_display_read_events(), no other thread will read from the fd and
+ * queue events in our queue.  If the call to
+ * wl_display_prepare_read() fails, we dispatch the pending events and
+ * try again until we're successful.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT int
+wl_display_prepare_read(struct wl_display *display)
+{
+	return wl_display_prepare_read_queue(display, &display->default_queue);
+}
+
+/** Release exclusive access to display file descriptor
+ *
+ * \param display The display context object
+ *
+ * This releases the exclusive access.  Useful for canceling the lock
+ * when a timed out poll returns fd not readable and we're not going
+ * to read from the fd anytime soon.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT void
+wl_display_cancel_read(struct wl_display *display)
+{
+	pthread_mutex_lock(&display->mutex);
+
+	display->reader_count--;
+	if (display->reader_count == 0) {
+		display->read_serial++;
+		pthread_cond_broadcast(&display->reader_cond);
+	}
+
+	pthread_mutex_unlock(&display->mutex);
+}
+
 /** Dispatch events in an event queue
  *
  * \param display The display context object
@@ -930,7 +1301,60 @@ WL_EXPORT int
 wl_display_dispatch_queue(struct wl_display *display,
 			  struct wl_event_queue *queue)
 {
-	return dispatch_queue(display, queue, 1);
+	struct pollfd pfd[2];
+	int ret;
+
+	pthread_mutex_lock(&display->mutex);
+
+	ret = dispatch_queue(display, queue);
+	if (ret == -1)
+		goto err_unlock;
+	if (ret > 0) {
+		pthread_mutex_unlock(&display->mutex);
+		return ret;
+	}
+
+	/* We ignore EPIPE here, so that we try to read events before
+	 * returning an error.  When the compositor sends an error it
+	 * will close the socket, and if we bail out here we don't get
+	 * a chance to process the error. */
+	ret = wl_connection_flush(display->connection);
+	if (ret < 0 && errno != EAGAIN && errno != EPIPE) {
+		display_fatal_error(display, errno);
+		goto err_unlock;
+	}
+
+	display->reader_count++;
+
+	pthread_mutex_unlock(&display->mutex);
+
+	pfd[0].fd = display->fd;
+	pfd[0].events = POLLIN;
+	do {
+		ret = poll(pfd, 1, -1);
+	} while (ret == -1 && errno == EINTR);
+
+	if (ret == -1) {
+		wl_display_cancel_read(display);
+		return -1;
+	}
+
+	pthread_mutex_lock(&display->mutex);
+
+	if (read_events(display) == -1)
+		goto err_unlock;
+
+	ret = dispatch_queue(display, queue);
+	if (ret == -1)
+		goto err_unlock;
+
+	pthread_mutex_unlock(&display->mutex);
+
+	return ret;
+
+ err_unlock:
+	pthread_mutex_unlock(&display->mutex);
+	return -1;
 }
 
 /** Dispatch pending events in an event queue
@@ -941,7 +1365,7 @@ wl_display_dispatch_queue(struct wl_display *display,
  *
  * Dispatch all incoming events for objects assigned to the given
  * event queue. On failure -1 is returned and errno set appropriately.
- * If there are no events queued, this functions return immediately.
+ * If there are no events queued, this function returns immediately.
  *
  * \memberof wl_display
  * \since 1.0.2
@@ -950,7 +1374,15 @@ WL_EXPORT int
 wl_display_dispatch_queue_pending(struct wl_display *display,
 				  struct wl_event_queue *queue)
 {
-	return dispatch_queue(display, queue, 0);
+	int ret;
+
+	pthread_mutex_lock(&display->mutex);
+
+	ret = dispatch_queue(display, queue);
+
+	pthread_mutex_unlock(&display->mutex);
+
+	return ret;
 }
 
 /** Process incoming events
@@ -969,7 +1401,8 @@ wl_display_dispatch_queue_pending(struct wl_display *display,
  * or not. For dispatching main queue events without blocking, see \ref
  * wl_display_dispatch_pending().
  *
- * \note Calling this makes the current thread the main one.
+ * \note Calling this will release the display file descriptor if this
+ * thread acquired it using wl_display_acquire_fd().
  *
  * \sa wl_display_dispatch_pending(), wl_display_dispatch_queue()
  *
@@ -978,9 +1411,7 @@ wl_display_dispatch_queue_pending(struct wl_display *display,
 WL_EXPORT int
 wl_display_dispatch(struct wl_display *display)
 {
-	display->display_thread = pthread_self();
-
-	return dispatch_queue(display, &display->queue, 1);
+	return wl_display_dispatch_queue(display, &display->default_queue);
 }
 
 /** Dispatch main queue events without reading from the display fd
@@ -1024,17 +1455,16 @@ wl_display_dispatch(struct wl_display *display)
 WL_EXPORT int
 wl_display_dispatch_pending(struct wl_display *display)
 {
-	display->display_thread = pthread_self();
-
-	return dispatch_queue(display, &display->queue, 0);
+	return wl_display_dispatch_queue_pending(display,
+						 &display->default_queue);
 }
 
-/** Retrieve the last error occurred on a display
+/** Retrieve the last error that occurred on a display
  *
  * \param display The display context object
- * \return The last error occurred on \c display or 0 if no error occurred
+ * \return The last error that occurred on \c display or 0 if no error occurred
  *
- * Return the last error occurred on the display. This may be an error sent
+ * Return the last error that occurred on the display. This may be an error sent
  * by the server or caused by the local client.
  *
  * \note Errors are \b fatal. If this function returns non-zero the display
@@ -1059,7 +1489,7 @@ wl_display_get_error(struct wl_display *display)
 /** Send all buffered requests on the display to the server
  *
  * \param display The display context object
- * \return The number of bytes send on success or -1 on failure
+ * \return The number of bytes sent on success or -1 on failure
  *
  * Send all buffered data on the client side to the server. Clients
  * should call this function before blocking. On success, the number
@@ -1164,7 +1594,10 @@ wl_proxy_get_class(struct wl_proxy *proxy)
 WL_EXPORT void
 wl_proxy_set_queue(struct wl_proxy *proxy, struct wl_event_queue *queue)
 {
-	proxy->queue = queue;
+	if (queue)
+		proxy->queue = queue;
+	else
+		proxy->queue = &proxy->display->default_queue;
 }
 
 WL_EXPORT void
