@@ -20,8 +20,8 @@
  * OF THIS SOFTWARE.
  */
 
-#ifndef WAYLAND_H
-#define WAYLAND_H
+#ifndef WAYLAND_SERVER_H
+#define WAYLAND_SERVER_H
 
 #ifdef  __cplusplus
 extern "C" {
@@ -77,13 +77,17 @@ int wl_event_loop_get_fd(struct wl_event_loop *loop);
 
 struct wl_client;
 struct wl_display;
-struct wl_seat;
-struct wl_pointer;
-struct wl_keyboard;
-struct wl_touch;
 struct wl_listener;
+struct wl_resource;
+struct wl_global;
 typedef void (*wl_notify_func_t)(struct wl_listener *listener, void *data);
 
+void wl_event_loop_add_destroy_listener(struct wl_event_loop *loop,
+					struct wl_listener * listener);
+struct wl_listener *wl_event_loop_get_destroy_listener(
+					struct wl_event_loop *loop,
+					wl_notify_func_t notify);
+
 struct wl_display *wl_display_create(void);
 void wl_display_destroy(struct wl_display *display);
 struct wl_event_loop *wl_display_get_event_loop(struct wl_display *display);
@@ -95,17 +99,20 @@ void wl_display_flush_clients(struct wl_display *display);
 typedef void (*wl_global_bind_func_t)(struct wl_client *client, void *data,
 				      uint32_t version, uint32_t id);
 
-struct wl_global *wl_display_add_global(struct wl_display *display,
-					const struct wl_interface *interface,
-					void *data,
-					wl_global_bind_func_t bind);
-
-void wl_display_remove_global(struct wl_display *display,
-			      struct wl_global *global);
-
 uint32_t wl_display_get_serial(struct wl_display *display);
 uint32_t wl_display_next_serial(struct wl_display *display);
 
+void wl_display_add_destroy_listener(struct wl_display *display,
+				     struct wl_listener *listener);
+struct wl_listener *wl_display_get_destroy_listener(struct wl_display *display,
+						    wl_notify_func_t notify);
+
+struct wl_global *wl_global_create(struct wl_display *display,
+				   const struct wl_interface *interface,
+				   int version,
+				   void *data, wl_global_bind_func_t bind);
+void wl_global_destroy(struct wl_global *global);
+
 struct wl_client *wl_client_create(struct wl_display *display, int fd);
 void wl_client_destroy(struct wl_client *client);
 void wl_client_flush(struct wl_client *client);
@@ -118,39 +125,115 @@ struct wl_listener *wl_client_get_destroy_listener(struct wl_client *client,
 						   wl_notify_func_t notify);
 
 struct wl_resource *
-wl_client_add_object(struct wl_client *client,
-		     const struct wl_interface *interface,
-		     const void *implementation, uint32_t id, void *data);
-struct wl_resource *
-wl_client_new_object(struct wl_client *client,
-		     const struct wl_interface *interface,
-		     const void *implementation, void *data);
-struct wl_resource *
 wl_client_get_object(struct wl_client *client, uint32_t id);
+void
+wl_client_post_no_memory(struct wl_client *client);
 
+/** \class wl_listener
+ *
+ * \brief A single listener for Wayland signals
+ *
+ * wl_listener provides the means to listen for wl_signal notifications. Many
+ * Wayland objects use wl_listener for notification of significant events like
+ * object destruction.
+ *
+ * Clients should create wl_listener objects manually and can register them as
+ * listeners to signals using #wl_signal_add, assuming the signal is
+ * directly accessible. For opaque structs like wl_event_loop, adding a
+ * listener should be done through provided accessor methods. A listener can
+ * only listen to one signal at a time.
+ *
+ * ~~~
+ * struct wl_listener your_listener;
+ *
+ * your_listener.notify = your_callback_method;
+ *
+ * \comment{Direct access}
+ * wl_signal_add(&some_object->destroy_signal, &your_listener);
+ *
+ * \comment{Accessor access}
+ * wl_event_loop *loop = ...;
+ * wl_event_loop_add_destroy_listener(loop, &your_listener);
+ * ~~~
+ *
+ * If the listener is part of a larger struct, #wl_container_of can be used
+ * to retrieve a pointer to it:
+ *
+ * \code
+ * void your_listener(struct wl_listener *listener, void *data)
+ * {
+ * 	struct your_data *data;
+ *
+ * 	your_data = wl_container_of(listener, data, your_member_name);
+ * }
+ * \endcode
+ *
+ * If you need to remove a listener from a signal, use #wl_list_remove.
+ *
+ * \code
+ * wl_list_remove(&your_listener.link);
+ * \endcode
+ *
+ * \sa wl_signal
+ */
 struct wl_listener {
 	struct wl_list link;
 	wl_notify_func_t notify;
 };
 
+/** \class wl_signal
+ *
+ * \brief A source of a type of observable event
+ *
+ * Signals are recognized points where significant events can be observed.
+ * Compositors as well as the server can provide signals. Observers are
+ * wl_listener's that are added through #wl_signal_add. Signals are emitted
+ * using #wl_signal_emit, which will invoke all listeners until that
+ * listener is removed by #wl_list_remove (or whenever the signal is
+ * destroyed).
+ *
+ * \sa wl_listener for more information on using wl_signal
+ */
 struct wl_signal {
 	struct wl_list listener_list;
 };
 
+/** Initialize a new \ref wl_signal for use.
+ *
+ * \param signal The signal that will be initialized
+ *
+ * \memberof wl_signal
+ */
 static inline void
 wl_signal_init(struct wl_signal *signal)
 {
 	wl_list_init(&signal->listener_list);
 }
 
+/** Add the specified listener to this signal.
+ *
+ * \param signal The signal that will emit events to the listener
+ * \param listener The listener to add
+ *
+ * \memberof wl_signal
+ */
 static inline void
 wl_signal_add(struct wl_signal *signal, struct wl_listener *listener)
 {
 	wl_list_insert(signal->listener_list.prev, &listener->link);
 }
 
+/** Gets the listener struct for the specified callback.
+ *
+ * \param signal The signal that contains the specified listener
+ * \param notify The listener that is the target of this search
+ * \return the list item that corresponds to the specified listener, or NULL
+ * if none was found
+ *
+ * \memberof wl_signal
+ */
 static inline struct wl_listener *
-wl_signal_get(struct wl_signal *signal, void *notify)
+wl_signal_get(struct wl_signal *signal, wl_notify_func_t notify)
 {
 	struct wl_listener *l;
 
@@ -161,6 +244,13 @@ wl_signal_get(struct wl_signal *signal, void *notify)
 	return NULL;
 }
 
+/** Emits this signal, notifying all registered listeners.
+ *
+ * \param signal The signal object that will emit the signal
+ * \param data The data that will be emitted with the signal
+ *
+ * \memberof wl_signal
+ */
 static inline void
 wl_signal_emit(struct wl_signal *signal, void *data)
 {
@@ -170,9 +260,19 @@ wl_signal_emit(struct wl_signal *signal, void *data)
 		l->notify(l, data);
 }
 
+typedef void (*wl_resource_destroy_func_t)(struct wl_resource *resource);
+
+#ifndef WL_HIDE_DEPRECATED
+
+struct wl_object {
+	const struct wl_interface *interface;
+	const void *implementation;
+	uint32_t id;
+};
+
 struct wl_resource {
 	struct wl_object object;
-	void (*destroy)(struct wl_resource *resource);
+	wl_resource_destroy_func_t destroy;
 	struct wl_list link;
 	struct wl_signal destroy_signal;
 	struct wl_client *client;
@@ -183,181 +283,34 @@ struct wl_buffer {
 	struct wl_resource resource;
 	int32_t width, height;
 	uint32_t busy_count;
-};
+} WL_DEPRECATED;
 
-struct wl_surface {
-	struct wl_resource resource;
-};
 
-struct wl_pointer_grab;
-struct wl_pointer_grab_interface {
-	void (*focus)(struct wl_pointer_grab *grab,
-		      struct wl_surface *surface,
-		      wl_fixed_t x,
-		      wl_fixed_t y);
-	void (*motion)(struct wl_pointer_grab *grab,
-		       uint32_t time,
-		       wl_fixed_t x,
-		       wl_fixed_t y);
-	void (*button)(struct wl_pointer_grab *grab,
-		       uint32_t time, uint32_t button, uint32_t state);
-};
+uint32_t
+wl_client_add_resource(struct wl_client *client,
+		       struct wl_resource *resource) WL_DEPRECATED;
 
-struct wl_pointer_grab {
-	const struct wl_pointer_grab_interface *interface;
-	struct wl_pointer *pointer;
-	struct wl_surface *focus;
-	wl_fixed_t x, y;
-};
+struct wl_resource *
+wl_client_add_object(struct wl_client *client,
+		     const struct wl_interface *interface,
+		     const void *implementation,
+		     uint32_t id, void *data) WL_DEPRECATED;
+struct wl_resource *
+wl_client_new_object(struct wl_client *client,
+		     const struct wl_interface *interface,
+		     const void *implementation, void *data) WL_DEPRECATED;
 
-struct wl_keyboard_grab;
-struct wl_keyboard_grab_interface {
-	void (*key)(struct wl_keyboard_grab *grab, uint32_t time,
-		    uint32_t key, uint32_t state);
-	void (*modifiers)(struct wl_keyboard_grab *grab, uint32_t serial,
-			  uint32_t mods_depressed, uint32_t mods_latched,
-			  uint32_t mods_locked, uint32_t group);
-};
+struct wl_global *
+wl_display_add_global(struct wl_display *display,
+		      const struct wl_interface *interface,
+		      void *data,
+		      wl_global_bind_func_t bind) WL_DEPRECATED;
 
-struct wl_keyboard_grab {
-	const struct wl_keyboard_grab_interface *interface;
-	struct wl_keyboard *keyboard;
-	struct wl_surface *focus;
-	uint32_t key;
-};
+void
+wl_display_remove_global(struct wl_display *display,
+			 struct wl_global *global) WL_DEPRECATED;
 
-struct wl_touch_grab;
-struct wl_touch_grab_interface {
-	void (*down)(struct wl_touch_grab *grab,
-			uint32_t time,
-			int touch_id,
-			wl_fixed_t sx,
-			wl_fixed_t sy);
-	void (*up)(struct wl_touch_grab *grab,
-			uint32_t time,
-			int touch_id);
-	void (*motion)(struct wl_touch_grab *grab,
-			uint32_t time,
-			int touch_id,
-			wl_fixed_t sx,
-			wl_fixed_t sy);
-};
-
-struct wl_touch_grab {
-	const struct wl_touch_grab_interface *interface;
-	struct wl_touch *touch;
-	struct wl_surface *focus;
-};
-
-struct wl_data_offer {
-	struct wl_resource resource;
-	struct wl_data_source *source;
-	struct wl_listener source_destroy_listener;
-};
-
-struct wl_data_source {
-	struct wl_resource resource;
-	struct wl_array mime_types;
-
-	void (*accept)(struct wl_data_source *source,
-		       uint32_t serial, const char *mime_type);
-	void (*send)(struct wl_data_source *source,
-		     const char *mime_type, int32_t fd);
-	void (*cancel)(struct wl_data_source *source);
-};
-
-struct wl_pointer {
-	struct wl_seat *seat;
-
-	struct wl_list resource_list;
-	struct wl_surface *focus;
-	struct wl_resource *focus_resource;
-	struct wl_listener focus_listener;
-	uint32_t focus_serial;
-	struct wl_signal focus_signal;
-
-	struct wl_pointer_grab *grab;
-	struct wl_pointer_grab default_grab;
-	wl_fixed_t grab_x, grab_y;
-	uint32_t grab_button;
-	uint32_t grab_serial;
-	uint32_t grab_time;
-
-	wl_fixed_t x, y;
-	struct wl_surface *current;
-	wl_fixed_t current_x, current_y;
-
-	uint32_t button_count;
-};
-
-struct wl_keyboard {
-	struct wl_seat *seat;
-
-	struct wl_list resource_list;
-	struct wl_surface *focus;
-	struct wl_resource *focus_resource;
-	struct wl_listener focus_listener;
-	uint32_t focus_serial;
-	struct wl_signal focus_signal;
-
-	struct wl_keyboard_grab *grab;
-	struct wl_keyboard_grab default_grab;
-	uint32_t grab_key;
-	uint32_t grab_serial;
-	uint32_t grab_time;
-
-	struct wl_array keys;
-
-	struct {
-		uint32_t mods_depressed;
-		uint32_t mods_latched;
-		uint32_t mods_locked;
-		uint32_t group;
-	} modifiers;
-};
-
-struct wl_touch {
-	struct wl_seat *seat;
-
-	struct wl_list resource_list;
-	struct wl_surface *focus;
-	struct wl_resource *focus_resource;
-	struct wl_listener focus_listener;
-	uint32_t focus_serial;
-	struct wl_signal focus_signal;
-
-	struct wl_touch_grab *grab;
-	struct wl_touch_grab default_grab;
-	wl_fixed_t grab_x, grab_y;
-	uint32_t grab_serial;
-	uint32_t grab_time;
-};
-
-struct wl_seat {
-	struct wl_list base_resource_list;
-	struct wl_signal destroy_signal;
-
-	struct wl_pointer *pointer;
-	struct wl_keyboard *keyboard;
-	struct wl_touch *touch;
-
-	uint32_t selection_serial;
-	struct wl_data_source *selection_data_source;
-	struct wl_listener selection_data_source_listener;
-	struct wl_signal selection_signal;
-
-	struct wl_list drag_resource_list;
-	struct wl_client *drag_client;
-	struct wl_data_source *drag_data_source;
-	struct wl_listener drag_data_source_listener;
-	struct wl_surface *drag_focus;
-	struct wl_resource *drag_focus_resource;
-	struct wl_listener drag_focus_listener;
-	struct wl_pointer_grab drag_grab;
-	struct wl_surface *drag_surface;
-	struct wl_listener drag_icon_listener;
-	struct wl_signal drag_icon_signal;
-};
+#endif
 
 /*
  * Post an event to the client's object referred to by 'resource'.
@@ -377,8 +330,12 @@ struct wl_seat {
  */
 void wl_resource_post_event(struct wl_resource *resource,
 			    uint32_t opcode, ...);
+void wl_resource_post_event_array(struct wl_resource *resource,
+				  uint32_t opcode, union wl_argument *args);
 void wl_resource_queue_event(struct wl_resource *resource,
 			     uint32_t opcode, ...);
+void wl_resource_queue_event_array(struct wl_resource *resource,
+				   uint32_t opcode, union wl_argument *args);
 
 /* msg is a printf format string, variable args are its args. */
 void wl_resource_post_error(struct wl_resource *resource,
@@ -388,98 +345,104 @@ void wl_resource_post_no_memory(struct wl_resource *resource);
 
 #include "wayland-server-protocol.h"
 
-uint32_t
-wl_client_add_resource(struct wl_client *client,
-		       struct wl_resource *resource);
-
 struct wl_display *
 wl_client_get_display(struct wl_client *client);
 
+struct wl_resource *
+wl_resource_create(struct wl_client *client,
+		   const struct wl_interface *interface,
+		   int version, uint32_t id);
+void
+wl_resource_set_implementation(struct wl_resource *resource,
+			       const void *implementation,
+			       void *data,
+			       wl_resource_destroy_func_t destroy);
+void
+wl_resource_set_dispatcher(struct wl_resource *resource,
+			   wl_dispatcher_func_t dispatcher,
+			   const void *implementation,
+			   void *data,
+			   wl_resource_destroy_func_t destroy);
+
 void
 wl_resource_destroy(struct wl_resource *resource);
-
+uint32_t
+wl_resource_get_id(struct wl_resource *resource);
+struct wl_list *
+wl_resource_get_link(struct wl_resource *resource);
+struct wl_resource *
+wl_resource_from_link(struct wl_list *resource);
+struct wl_resource *
+wl_resource_find_for_client(struct wl_list *list, struct wl_client *client);
+struct wl_client *
+wl_resource_get_client(struct wl_resource *resource);
 void
-wl_seat_init(struct wl_seat *seat);
-
-void
-wl_seat_release(struct wl_seat *seat);
-
-void
-wl_seat_set_pointer(struct wl_seat *seat, struct wl_pointer *pointer);
-void
-wl_seat_set_keyboard(struct wl_seat *seat, struct wl_keyboard *keyboard);
-void
-wl_seat_set_touch(struct wl_seat *seat, struct wl_touch *touch);
-
-void
-wl_pointer_init(struct wl_pointer *pointer);
-void
-wl_pointer_release(struct wl_pointer *pointer);
-void
-wl_pointer_set_focus(struct wl_pointer *pointer, struct wl_surface *surface,
-		     wl_fixed_t sx, wl_fixed_t sy);
-void
-wl_pointer_start_grab(struct wl_pointer *pointer,
-		      struct wl_pointer_grab *grab);
-void
-wl_pointer_end_grab(struct wl_pointer *pointer);
-
-void
-wl_keyboard_init(struct wl_keyboard *keyboard);
-void
-wl_keyboard_release(struct wl_keyboard *keyboard);
-void
-wl_keyboard_set_focus(struct wl_keyboard *keyboard, struct wl_surface *surface);
-void
-wl_keyboard_start_grab(struct wl_keyboard *device,
-		       struct wl_keyboard_grab *grab);
-void
-wl_keyboard_end_grab(struct wl_keyboard *keyboard);
-
-void
-wl_touch_init(struct wl_touch *touch);
-void
-wl_touch_release(struct wl_touch *touch);
-void
-wl_touch_start_grab(struct wl_touch *device,
-		struct wl_touch_grab *grab);
-void
-wl_touch_end_grab(struct wl_touch *touch);
-
-void
-wl_data_device_set_keyboard_focus(struct wl_seat *seat);
-
+wl_resource_set_user_data(struct wl_resource *resource, void *data);
+void *
+wl_resource_get_user_data(struct wl_resource *resource);
 int
-wl_data_device_manager_init(struct wl_display *display);
-
+wl_resource_get_version(struct wl_resource *resource);
+void
+wl_resource_set_destructor(struct wl_resource *resource,
+			   wl_resource_destroy_func_t destroy);
+int
+wl_resource_instance_of(struct wl_resource *resource,
+			const struct wl_interface *interface,
+			const void *implementation);
 
 void
-wl_seat_set_selection(struct wl_seat *seat,
-		      struct wl_data_source *source, uint32_t serial);
+wl_resource_add_destroy_listener(struct wl_resource *resource,
+				 struct wl_listener * listener);
+struct wl_listener *
+wl_resource_get_destroy_listener(struct wl_resource *resource,
+				 wl_notify_func_t notify);
 
+#define wl_resource_for_each(resource, list)					\
+	for (resource = 0, resource = wl_resource_from_link((list)->next);	\
+	     wl_resource_get_link(resource) != (list);				\
+	     resource = wl_resource_from_link(wl_resource_get_link(resource)->next))
+
+#define wl_resource_for_each_safe(resource, tmp, list)					\
+	for (resource = 0, tmp = 0,							\
+	     resource = wl_resource_from_link((list)->next),	\
+	     tmp = wl_resource_from_link((list)->next->next);	\
+	     wl_resource_get_link(resource) != (list);				\
+	     resource = tmp,							\
+	     tmp = wl_resource_from_link(wl_resource_get_link(resource)->next))
+
+struct wl_shm_buffer;
+
+void
+wl_shm_buffer_begin_access(struct wl_shm_buffer *buffer);
+
+void
+wl_shm_buffer_end_access(struct wl_shm_buffer *buffer);
+
+struct wl_shm_buffer *
+wl_shm_buffer_get(struct wl_resource *resource);
 
 void *
-wl_shm_buffer_get_data(struct wl_buffer *buffer);
+wl_shm_buffer_get_data(struct wl_shm_buffer *buffer);
 
 int32_t
-wl_shm_buffer_get_stride(struct wl_buffer *buffer);
+wl_shm_buffer_get_stride(struct wl_shm_buffer *buffer);
 
 uint32_t
-wl_shm_buffer_get_format(struct wl_buffer *buffer);
+wl_shm_buffer_get_format(struct wl_shm_buffer *buffer);
 
 int32_t
-wl_shm_buffer_get_width(struct wl_buffer *buffer);
+wl_shm_buffer_get_width(struct wl_shm_buffer *buffer);
 
 int32_t
-wl_shm_buffer_get_height(struct wl_buffer *buffer);
-
-int
-wl_buffer_is_shm(struct wl_buffer *buffer);
+wl_shm_buffer_get_height(struct wl_shm_buffer *buffer);
 
 int
 wl_display_init_shm(struct wl_display *display);
 
-struct wl_buffer *
+uint32_t *
+wl_display_add_shm_format(struct wl_display *display, uint32_t format);
+
+struct wl_shm_buffer *
 wl_shm_buffer_create(struct wl_client *client,
 		     uint32_t id, int32_t width, int32_t height,
 		     int32_t stride, uint32_t format);
